#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <png.h>

#define NBVERTICES 35947L

typedef enum Axis { x, y, z } Axis;
#define right x
#define red x
#define up y
#define green y
#define front z
#define blue z
#define width x
#define height y

double* vertices;
png_byte color_type = PNG_COLOR_TYPE_RGB;
png_byte bit_depth = 8;
png_structp png_ptr;
png_infop info_ptr;
png_bytep * row_pointers;

char* filename[8] = {
  "front.png", "right.png", "top.png", "diag.png", "zoom.png",
  "rear.png", "left.png", "bottom.png"};

typedef struct Vec3 {
  double c[3];
} Vec3;

typedef struct Camera {
  double fov;
  double center[2];
  double pos[3];
  double rot[3][3];  // rot[right/up/front][x/y/z]
} Camera;

typedef struct Screen {
  double size[2];
} Screen;

Screen screen;
Camera camera;

Vec3 Project(Camera camera, Screen screen, Vec3 point) {
  double focal = screen.size[width] / (2.0 * tan(camera.fov * 0.5));
  double v[3] = {
    point.c[x] - camera.pos[x],
    point.c[y] - camera.pos[y],
    point.c[z] - camera.pos[z]};
  double u[3] = {
    camera.rot[right][x] * v[x] +
      camera.rot[right][y] * v[y] +
      camera.rot[right][z] * v[z],
    camera.rot[up][x] * v[x] +
      camera.rot[up][y] * v[y] +
      camera.rot[up][z] * v[z],
    camera.rot[front][x] * v[x] +
      camera.rot[front][y] * v[y] +
      camera.rot[front][z] * v[z]};
  Vec3 pScreen = {.c={
    camera.center[x] + focal * u[x] / u[z],
    camera.center[y] + focal * u[y] / u[z],
    u[z]}};
  return pScreen;
}

void CheckProj(int iTest) {

  FILE *fp;
  fp = fopen(filename[iTest], "wb");
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  info_ptr = png_create_info_struct(png_ptr);
  png_init_io(png_ptr, fp);
  png_set_IHDR(
    png_ptr, info_ptr, screen.size[width], screen.size[height],
    bit_depth, color_type, PNG_INTERLACE_NONE,
    PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  png_write_info(png_ptr, info_ptr);

  for (
    unsigned long iVertex =0;
    iVertex < NBVERTICES;
    ++iVertex) {

    Vec3 pWorld = {
      .c={
        vertices[iVertex * 3L], 
        vertices[iVertex * 3L + 1], 
        vertices[iVertex * 3L + 2]}};
    Vec3 pScreen = Project(camera, screen, pWorld);
    if (pScreen.c[x] > 0.0 && pScreen.c[y] > 0.0 && pScreen.c[z] > 0.0 &&
      pScreen.c[x] < (screen.size[width] - 1.0) &&
      pScreen.c[y] < (screen.size[height] - 1.0)) {
      png_byte* row = row_pointers[(int)screen.size[height] - 1 - (int)pScreen.c[y]];
      png_byte* ptr = &(row[(int)pScreen.c[x]*3]);
      ptr[red] = 0;
      ptr[green] = 0;
      ptr[blue] = 255;
    }

  }

  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, NULL);
  for (int v = 0; v < screen.size[height]; v++) free(row_pointers[v]);
  free(row_pointers);
  fclose(fp);

}

void LoadBunny() {

  vertices = malloc(sizeof(double) * 3L * NBVERTICES);

  // bun_zipper.ply downloaded from the Stanford 3D scanning repository
  // http://graphics.stanford.edu/data/3Dscanrep/
  // 
  FILE* fp = fopen("./bun_zipper.ply", "r");

  for (int i = 12; i--;) {
    char buffer[1000];
    char* ret = fgets(buffer, 1000, fp);
  }

  for (unsigned long iVertex = 0; iVertex < NBVERTICES; ++iVertex) {
    double buff;
    int ret = fscanf(fp,
        "%lf %lf %lf %lf %lf\n",
        vertices + 3L * iVertex,
        vertices + 3L * iVertex + 1L,
        vertices + 3L * iVertex + 2L,
        &buff,
        &buff);
    vertices[3L * iVertex + 2L] *= -1.0;
  }

  fclose(fp);

}

void RenderPov(int iTest) {

  FILE* fpPov =
    fopen(
      "bunny.pov",
      "w");

  fprintf(
    fpPov,
    "#declare bunny = union {\n");

  for (
    unsigned long iVertex =0;
    iVertex < NBVERTICES;
    ++iVertex) {

    fprintf(
      fpPov,
      "sphere {<%lf,%lf,%lf>, 0.0002}\n",
      vertices[iVertex * 3L],
      vertices[iVertex * 3L + 1],
      vertices[iVertex * 3L + 2]);

  }

  fprintf(
    fpPov,
    "}\n");
  fprintf(
    fpPov,
"#version 3.7;\n"
"background { color rgb 1 }\n"
"camera {\n");
  fprintf(
    fpPov,
"  location <%lf, %lf, %lf>\n"
"  look_at <%lf, %lf, %lf>\n"
"  angle %lf\n",
    camera.pos[x],
    camera.pos[y],
    camera.pos[z],
    camera.pos[x] + camera.rot[front][x],
    camera.pos[y] + camera.rot[front][y],
    camera.pos[z] + camera.rot[front][z],
    camera.fov * (180.0 / M_PI));
  fprintf(
    fpPov,
"}\n"
"object {\n"
"  bunny\n"
"  texture {pigment {color rgb x} finish {ambient 1.0 diffuse 1.0}}\n"
"}\n");

  fclose(fpPov);

  // POV-Ray rendering engine available from http://www.povray.org/
  char cmd[1000];
  sprintf(cmd, "povray +Ibunny.pov +W%d +H%d +O%s",
    (int)screen.size[width], (int)screen.size[height], filename[iTest]);
  system(cmd);

  FILE *fp;
  fp = fopen(filename[iTest], "rb");

  char header[8];
  fread(header, 1, 8, fp);
  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  info_ptr = png_create_info_struct(png_ptr);
  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, 8);
  png_read_info(png_ptr, info_ptr);
  color_type = png_get_color_type(png_ptr, info_ptr);
  bit_depth = png_get_bit_depth(png_ptr, info_ptr);
  int number_of_passes = png_set_interlace_handling(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
  row_pointers =
    (png_bytep*)malloc(sizeof(png_bytep) * (int)screen.size[height]);
  for (int v=0; v<(int)screen.size[height]; v++)
    row_pointers[v] = (png_byte*) malloc(png_get_rowbytes(png_ptr,info_ptr));
  png_read_image(png_ptr, row_pointers);
  fclose(fp);

}

double rad(double d) {
  return d / (180.0 / M_PI);
}

int main() {

  for (int iTest = 0; iTest < 8; ++ iTest) {

    LoadBunny();

    if (iTest == 0) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.0, 0.1, -0.5},
        .rot = {
          {1.0, 0.0, 0.0}, // right
          {0.0, 1.0, 0.0}, // up
          {0.0, 0.0, 1.0}} // front
      };
    } else if (iTest == 1) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.5, 0.1, 0.0},
        .rot = {
          {0.0, 0.0, 1.0}, // right
          {0.0, 1.0, 0.0}, // up
          {-1.0, 0.0, 0.0}} // front
      };
    } else if (iTest == 2) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.0, 0.5, 0.0},
        .rot = {
          {1.0, 0.0, 0.0}, // right
          {0.0, 0.0, 1.0}, // up
          {0.0, -1.0, 0.0}} // front
      };
    } else if (iTest == 3) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.25, 0.5, 0.25},
        .rot = {
          {-0.707107, 0.000000, 0.707107}, // right
          {-0.500000, 0.707107, -0.500000}, // up
          {-0.500000, -0.707107, -0.500000}} // front
      };
    } else if (iTest == 4) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(30.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.25, 0.5, 0.25},
        .rot = {
          {-0.707107, 0.000000, 0.707107}, // right
          {-0.500000, 0.707107, -0.500000}, // up
          {-0.500000, -0.707107, -0.500000}} // front
      };
    } else if (iTest == 5) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.0, 0.1, 0.5},
        .rot = {
          {-1.0, 0.0, 0.0}, // right
          {0.0, 1.0, 0.0}, // up
          {0.0, 0.0, -1.0}} // front
      };
    } else if (iTest == 6) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {-0.5, 0.1, 0.0},
        .rot = {
          {0.0, 0.0, -1.0}, // right
          {0.0, 1.0, 0.0}, // up
          {1.0, 0.0, 0.0}} // front
      };
    } else if (iTest == 7) {
      screen = (Screen){ .size = {800., 600.}};
      camera = (Camera){
        .fov = rad(45.0),
        .center = {screen.size[width] * 0.5, screen.size[height] * 0.5},
        .pos = {0.0, -0.5, 0.0},
        .rot = {
          {1.0, 0.0, 0.0}, // right
          {0.0, 0.0, -1.0}, // up
          {0.0, 1.0, 0.0}} // front
      };
    }

    RenderPov(iTest);
    CheckProj(iTest);

  }

  return 0;
}
